import json
import logging
import pathlib
import pandas as pd
import numpy as np

from PySide6.QtCore import QRunnable, Slot

from fixtracks.utils.signals import ProducerSignals


class Converter(QRunnable):
    def __init__(self, infile, outfile) -> None:
        super().__init__()
        self._signals = ProducerSignals()
        self._infile = infile
        self._outfile = outfile
        self._ok = True
        self._stopRequest = False
        self._checkfiles()

    def _checkfiles(self):
        for file in [self._infile, self._outfile]:
            if file is None:
                self._signals.error.emit(f"Converter.__init__: {file} is None!")
                self._ok = False
            if not pathlib.Path.exists(pathlib.Path(file)):
                self._signals.error.emit("Converter.__init__: File {file} does not exist!")
                self._ok = False

    @Slot()
    def stop_request(self):
        self._stopRequest = True

    @Slot()
    def run(self):
        logging.info(f"Converting input file {self._infile} to {self._outfile}")
        self._signals.progress.emit(0.0)
        tracking_data = {"frame": [], "name": [], "cls": [],
                        "confidence": [], "visible": [] ,
                        "track_id": [],
                        }
        key_dict = {f"key_{i}": [] for i in range(6)}
        box_labels =  ["x1", "y1", "x2", "y2"]
        box_dict = {f"box_{i}": [] for i in box_labels}
        tracking_data.update(key_dict)
        tracking_data.update(box_dict)
        logging.info("Reading infile")
        with open(self._infile, "r", encoding="utf-8") as f:
            data = json.load(f)
        progress = 0.1
        self._signals.progress.emit(progress)
        logging.info("Reading infile done")
        interval = int(len(data) // 80)

        for i, d in enumerate(data):  # each of the frames
            if len(d["data"]) == 0:
                continue
            if i % interval == 0:
                progress += 0.8/80
                self._signals.progress.emit(np.round(progress, 1))
                if self._stopRequest:
                    self._signals.error.emit("Conversion cancelled!")
                    break
            for dd in d["data"]:  # each of the found objects, i.e. fish
                key_x = [float(x) for x in list(np.round(dd["keypoints"]["x"], 2))]
                key_y = [float(y) for y in list(np.round(dd["keypoints"]["y"], 2))]
                key_dict = {f"key_{i}": [v] for i, v in enumerate(zip(key_x, key_y))}
                visible = np.round(dd["keypoints"]["visible"], 3)
                tracking_data["frame"].append(d["frame"])
                tracking_data["name"].append(dd["name"])
                tracking_data["cls"].append(dd["class"])
                tracking_data["visible"].append(visible)
                tracking_data["confidence"].append(float(np.round(dd["confidence"], 3)))
                tracking_data["track_id"].append(dd["track_id"] if "track_id" in dd.keys() else -1)
                for bk, bl in zip(box_dict.keys(), box_labels):
                    tracking_data[bk].append(dd["box"][bl])
                for i, kd in enumerate(key_dict.keys()):
                    tracking_data[kd].append((key_x[i], key_y[i]))

        if self._stopRequest:
            self._signals.error.emit("Conversion cancelled!")
            return
        df = pd.DataFrame(tracking_data)
        df.to_csv(self._outfile, sep=";")
        self._signals.progress.emit(1.0)
        self._signals.finished.emit(True and (not self._stopRequest))

    @property
    def signals(self):
        return self._signals

# import cv2
# def topandas(filename):
#     tracking_data = {"frame": [], "name": [], "cls": [],
#                      "confidence": [], "visible": [] ,
#                      "track_id": [],
#                      }
#     key_dict = {f"key_{i}": [] for i in range(6)}
#     box_labels =  ["x1", "y1", "x2", "y2"]
#     box_dict = {f"box_{i}": [] for i in box_labels}
#     tracking_data.update(key_dict)
#     tracking_data.update(box_dict)

#     with open(filename, "r", encoding="utf-8") as f:
#         data = json.load(f)
#     for d in data:  # each of the frames
#         if len(d["data"]) == 0:
#             continue
#         for dd in d["data"]:  # each of the found objects, i.e. fish
#             key_x = [float(x) for x in list(np.round(dd["keypoints"]["x"], 2))]
#             key_y = [float(y) for y in list(np.round(dd["keypoints"]["y"], 2))]
#             key_dict = {f"key_{i}": [v] for i, v in enumerate(zip(key_x, key_y))}
#             visible = np.round(dd["keypoints"]["visible"], 3)
#             tracking_data["frame"].append(d["frame"])
#             tracking_data["name"].append(dd["name"])
#             tracking_data["cls"].append(dd["class"])
#             tracking_data["visible"].append(visible)
#             tracking_data["confidence"].append(float(np.round(dd["confidence"], 3)))
#             tracking_data["track_id"].append(dd["track_id"] if "track_id" in dd.keys() else -1)
#             for bk, bl in zip(box_dict.keys(), box_labels):
#                 tracking_data[bk].append(dd["box"][bl])
#             for i, kd in enumerate(key_dict.keys()):
#                 tracking_data[kd].append((key_x[i], key_y[i]))

#     df = pd.DataFrame(tracking_data)
#     return df


# import matplotlib.pyplot as plt

# from IPython import embed

# def show_video(filename):
#     cap = cv2.VideoCapture('2024.11.13_0_converted_right_undistorted_fixed.mp4')
#     count = 0
#     while cap.isOpened():
#         ret, frame = cap.read()
#         # if frame is read correctly ret is True
#         if not ret:
#             print("Can't receive frame (stream end?). Exiting ...")
#             break
#         count += 1
#         if count < 2000:
#             continue
#         gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
#         print(count)
#         cv2.imshow('frame', gray[::2,::2])
#         if cv2.waitKey(1) == ord('q'):
#             break
#     cap.release()
#     cv2.destroyAllWindows()


# Test code
def main():
    # import left and right tracks into a pandas dataframe
    infile = "../../data/2024.11.13_0_converted_right_undistorted_fixed_pose.json"
    outfile = "../../data/right_tracks.csv"
    converter = Converter(infile, outfile)
    converter.run()

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, force=True)
    main()
