import logging

from PySide6.QtWidgets import QWidget, QStackedLayout, QSizePolicy, QVBoxLayout, QGridLayout, QSpacerItem
from PySide6.QtCore import Qt

from fixtracks.widgets.tracks import FixTracks
from fixtracks.widgets.detectionmerge import MergeDetections
from fixtracks.widgets.taskwidget import TasksWidget, Task
from fixtracks.widgets.converter import Json2PandasConverter
from fixtracks.widgets.taskcard import TaskCard


class CentralWidget(QWidget):

    def __init__(self, darkmode, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self._folderopen = False
        self._tw = TasksWidget(darkmode)
        self._tw.clicked.connect(self._select_task)
        self._tw.opened.connect(self.on_open)

        self._mergewidget = MergeDetections(self)
        self._mergewidget.back.connect(self._on_back)

        self._convertwidget = Json2PandasConverter(self)
        self._convertwidget.back.connect(self._on_back)

        self._trackwidget = FixTracks(self)
        self._trackwidget.back.connect(self._on_back)
        version = "dark" if darkmode else "light"
        convert_task = TaskCard(Task.CONVERT.value, f":/icons/convert_{version}",
                                "Convert JSON pose files to Pandas data frames")
        convert_task.clicked.connect(lambda: self._select_task(Task.CONVERT))
        merge_task = TaskCard(Task.MERGE.value, f":/icons/merge_{version}", "Merge detections from two cameras into the same dataset")
        merge_task.clicked.connect(lambda: self._select_task(Task.MERGE))
        track_task = TaskCard(Task.TRACKS.value, f":/icons/tracks_{version}", "Fix tracking of individuals, reassign track ids.")
        track_task.clicked.connect(lambda: self._select_task(Task.TRACKS))

        grid = QGridLayout()
        grid.addWidget(convert_task, 0, 0, alignment=Qt.AlignmentFlag.AlignCenter)
        grid.addWidget(merge_task, 1, 0, alignment=Qt.AlignmentFlag.AlignCenter)
        grid.addWidget(track_task, 0, 1, alignment=Qt.AlignmentFlag.AlignCenter)
        grid.setAlignment(Qt.AlignmentFlag.AlignCenter)
        l = QVBoxLayout()
        l.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        l.addItem(QSpacerItem(100, 100, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))
        l.addLayout(grid)
        l.addItem(QSpacerItem(100, 100, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))
        taskgrid = QWidget()
        taskgrid.setLayout(l)

        layout = QStackedLayout()
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._btnindex = layout.addWidget(self._tw)
        self._tasksindex = layout.addWidget(taskgrid)
        self._convertindex = layout.addWidget(self._convertwidget)
        self._mergeindex = layout.addWidget(self._mergewidget)
        self._trackindex = layout.addWidget(self._trackwidget)
        self.setLayout(layout)

    def on_open(self):
        self.layout().setCurrentIndex(self._tasksindex)
        self._folderopen = True

    def _select_task(self, task):
        logging.debug("Centralwidget: Selected task: %s", task.value)
        if task == Task.MERGE:
            self.layout().setCurrentIndex(self._mergeindex)
            self._mergewidget.fileList = self._tw.fileList
        elif task == Task.TRACKS:
            self.layout().setCurrentIndex(self._trackindex)
            self._trackwidget.fileList = self._tw.fileList
        elif task == Task.CONVERT:
            self.layout().setCurrentIndex(self._convertindex)
        else:
            if self._folderopen:
                self.layout().setCurrentIndex(self._tasksindex)
            else:
                self.layout().setCurrentIndex(self._btnindex)
            logging.warning("Centralwidget: got invalid task request: %s", task)

    def _on_back(self):
        self.layout().setCurrentIndex(self._tasksindex)

    @property
    def toolbarActions(self):
        return self._tw.toolbarActions

    @property
    def menuActions(self):
        return self._tw.menuActions


def main():
    from PySide6.QtWidgets import QApplication
    app = QApplication([])
    window = QWidget()
    taskcard = TaskCard("test", "../icons/merge_light.png", "do something with manyt words, too many to fit in one line")
    layout = QVBoxLayout()
    layout.addWidget(taskcard)

    window.setLayout(layout)
    window.show()
    app.exec()

if __name__ == "__main__":
    main()